"""

"""
import socket
from abc import abstractmethod
from collections import namedtuple
from typing import Tuple, Protocol, Optional

#** Variables **#
__all__ = [
    'modify_socket',

    'RawAddr',
    'Address', 
    'Writer', 
    'Session',
]

#: raw address tuple
RawAddr = Tuple[str, int]

#: named address tuple
Address = namedtuple('Address', ('host', 'port'))

#** Functions **#

def modify_socket(
    sock:      socket.socket, 
    timeout:   Optional[int], 
    interface: Optional[bytes]
) -> socket.socket:
    """
    modify the socket object w/ the following options
    """
    if timeout is not None:
        sock.settimeout(timeout)
    if interface is not None:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BINDTODEVICE, interface)
    return sock 

#** Classes **#

class Writer(Protocol):
    """
    abstract data-writing implementation for single server connection
    """

    @abstractmethod
    def write(self, data: bytes):
        raise NotImplementedError
    
    @abstractmethod
    def close(self):
        raise NotImplementedError

    @abstractmethod
    def is_closing(self) -> bool:
        raise NotImplementedError

class Session(Protocol):
    """
    abstract session-manager implemention for single server connection
    """
    
    @abstractmethod
    def connection_made(self, addr: Address, writer: Writer):
       raise NotImplementedError

    @abstractmethod
    def data_recieved(self, data: bytes):
        raise NotImplementedError

    @abstractmethod
    def connection_lost(self, err: Optional[Exception]):
        raise NotImplementedError



"""
Unify SocketServer Implementations based on Session Model
"""
import asyncio
from ssl import SSLContext
from typing import Type, Optional

from .abc import RawAddr, Address, Writer, Session
from .transport import UdpProtocol, TcpProtocol
from .threading import UdpThreadServer, TcpThreadServer

#** Variables **#
__all__ = [
    'listen_udp_async',
    'listen_tcp_async',
    'listen_udp_threaded',
    'listen_tcp_threaded',

    'Address',
    'Writer',
    'Session',
]

#** Functions **#

async def listen_udp_async(
    address:         RawAddr, 
    factory:         Type[Session],
    *args,
    timeout:         Optional[int]   = None,
    interface:       Optional[bytes] = None,
    reuse_port:      bool            = False,
    allow_broadcast: bool            = False,
    **kwargs,
):
    """
    :param address:         host/port of server
    :param factory:         type factory for server request handler
    :param args:            positional args to pass to the session factory
    :param timeout:         max socket lifetime duration timeout if configured
    :param interface:       interface to bind server socket to
    :param reuse_port:      allow reuse of same port when enabled
    :param allow_broadcast: allow for udp broadcast messages when enabled
    :param kwargs:          keyword arguments to pass to session factory
    """
    # spawn protocol factory and test session generation
    loop = asyncio.get_running_loop()
    func = lambda: UdpProtocol(factory, args, kwargs, timeout, interface)
    func().test_factory()
    # spawn server instance
    tport, _ = await loop.create_datagram_endpoint(func, address, 
        reuse_port=reuse_port, allow_broadcast=allow_broadcast)
    # run server forever
    try:
        while True:
            await asyncio.sleep(3600)
    finally:
        tport.close()

async def listen_tcp_async(
    address:    RawAddr, 
    factory:    Type[Session],
    *args,
    timeout:    Optional[int]        = None,
    interface:  Optional[bytes]      = None,
    reuse_port: bool                 = False,
    backlog:    int                  = 100,
    ssl:        Optional[SSLContext] = None,
    **kwargs,
):
    """
    :param address:         host/port of server
    :param factory:         type factory for server request handler
    :param args:            positional args to pass to the session factory
    :param timeout:         max socket lifetime duration timeout if configured
    :param interface:       interface to bind server socket to
    :param reuse_port:      allow reuse of same port when enabled
    :param backlog:         number of allowed and backloged async requests
    :param ssl:             TLS context to encrypt socket communications
    :param kwargs:          keyword arguments to pass to session factory
    """
    # spawn protocol factory and test session generation
    loop = asyncio.get_running_loop()
    func = lambda: TcpProtocol(factory, args, kwargs, timeout, interface)
    func().test_factory()
    # spawn server instance
    host, port = address
    server = await loop.create_server(func, host, port, 
        reuse_port=reuse_port, backlog=backlog, ssl=ssl)
    # run server forever
    async with server:
        await server.serve_forever()

def listen_udp_threaded(
    address:         RawAddr, 
    factory:         Type[Session],
    *args,
    timeout:         Optional[int]   = None,
    interface:       Optional[bytes] = None,
    reuse_port:      bool            = False,
    allow_broadcast: bool            = False,
    **kwargs,
):
    """
    :param address:         host/port of server
    :param factory:         type factory for server request handler
    :param args:            positional args to pass to the session factory
    :param timeout:         max socket lifetime duration timeout if configured
    :param interface:       interface to bind server socket to
    :param reuse_port:      allow reuse of same port when enabled
    :param allow_broadcast: allow for udp broadcast messages when enabled
    :param kwargs:          keyword arguments to pass to session factory
    """
    server = UdpThreadServer(
        address=address, 
        factory=factory, 
        args=args, 
        kwargs=kwargs, 
        timeout=timeout, 
        interface=interface, 
        reuse_port=reuse_port, 
        allow_broadcast=allow_broadcast,
    )
    with server:
        server.serve_forever()

def listen_tcp_threaded(
    address:    RawAddr,
    factory:    Type[Session],
    *args,
    timeout:    Optional[int]        = None,
    interface:  Optional[bytes]      = None,
    reuse_port: bool                 = False,
    ssl:        Optional[SSLContext] = None,
    blocksize:  int                  = 8192,
    **kwargs,
):
    """
    :param address:         host/port of server
    :param factory:         type factory for server request handler
    :param args:            positional args to pass to the session factory
    :param timeout:         max socket lifetime duration timeout if configured
    :param interface:       interface to bind server socket to
    :param reuse_port:      allow reuse of same port when enabled
    :param backlog:         number of allowed and backloged async requests
    :param ssl:             TLS context to encrypt socket communications
    :param kwargs:          keyword arguments to pass to session factory
    """
    factory.test_factory()
    server = TcpThreadServer(
        address=address, 
        factory=factory, 
        args=args, 
        kwargs=kwargs, 
        timeout=timeout, 
        interface=interface, 
        reuse_port=reuse_port, 
        ssl=ssl,
        blocksize=blocksize,
    )
    with server:
        server.serve_forever()



"""
Threading Implementations of session-based servers
"""
import socket
import socketserver
from ssl import SSLContext, wrap_socket
from dataclasses import dataclass, field
from typing import Type, Optional, Dict, Any, ClassVar

from .abc import RawAddr, Address, Writer, Session, modify_socket

#** Variables **#
__all__ = ['UdpThreadServer', 'TcpThreadServer']

#** Functions **#

def new_handler(
    base:      Type['BaseRequestHandler'],
    factory:   Type[Session],
    args:      tuple           = (),
    kwargs:    Dict[str, Any]  = {},
    timeout:   Optional[int]   = None,
    interface: Optional[bytes] = None,
    blocksize: int             = 8192
) -> Type['BaseRequestHandler']:
    """
    spawn new request-handler class w/ configured settings
    """
    # test factory generation
    factory(*args, **kwargs)
    # generate new request handler
    name = f'{base.__name__}Instance'
    return type(name, (base, ), dict(
        factory=factory,
        args=args,
        kwargs=kwargs,
        timeout=timeout,
        interface=interface,
        blocksize=blocksize,
    ))

#** Classes **#

@dataclass
class UdpWriter(Writer):
    addr: Address
    sock: socket.socket
    closing: bool = False
    
    def write(self, data: bytes, addr: Optional[Address] = None):
        self.sock.sendto(data, addr or self.addr)

    def close(self):
        self.sock.close()
        self.closing = True

    def is_closing(self) -> bool:
        return self.closing

@dataclass
class TcpWriter(Writer):
    sock: socket.socket
    closing: bool = False

    def write(self, data: bytes):
        self.sock.sendall(data)

    def close(self):
        self.closing = True
        self.sock.close()

    def is_closing(self) -> bool:
        return self.closing

class BaseRequestHandler(socketserver.BaseRequestHandler):
    factory:   Type[Session]
    args:      tuple           
    kwargs:    Dict[str, Any]
    timeout:   Optional[int]
    interface: Optional[bytes]
    blocksize: int

    def setup(self):
        """configure and generate session w/ information collected"""
        self.addr:   Address = Address(*self.client_address)
        self.sock:   socket.socket
        self.writer: Writer
        self.error:  Optional[Exception] = None
        modify_socket(self.sock, self.timeout, self.interface)
        # spawn session object
        self.session = self.factory(*self.args, **self.kwargs)
        self.session.connection_made(self.addr, self.writer)
    
    def finish(self):
        """notify that connection disconnected"""
        self.session.connection_lost(self.error)

class UdpHandler(BaseRequestHandler):
    
    def setup(self):
        """handle connection spawn"""
        self.addr    = Address(*self.client_address) 
        self.sock    = self.request[1]
        self.writer: UdpWriter = UdpWriter(self.addr, self.sock)
        super().setup()

    def handle(self):
        """handle single inbound udp packet"""
        try:
            data = self.request[0]
            self.session.data_recieved(data)
        except socket.error as e:
            self.error = e
            if not self.writer.closing:
                self.writer.close()

class TcpHandler(BaseRequestHandler):
    
    def setup(self):
        """handle setup of server"""
        self.sock = self.request
        self.writer: TcpWriter = TcpWriter(self.sock)
        super().setup()

    def handle(self):
        """handle subsequent reads of inbound data"""
        while not self.writer.closing:
            try:
                data = self.sock.recv(self.blocksize)
                if not data:
                    break
                self.session.data_recieved(data)
            except socket.error as e:
                self.error = e
                if not self.writer.closing:
                    self.writer.close()
                break

@dataclass
class BaseThreadServer(socketserver.ThreadingMixIn):
    server:    ClassVar[Type[socketserver.BaseServer]]
    handler:   ClassVar[Type[BaseRequestHandler]]

    address:    RawAddr
    factory:    Type[Session]
    args:       tuple           = field(default_factory=tuple)
    kwargs:     Dict[str, Any]  = field(default_factory=dict)
    timeout:    Optional[int]   = None
    interface:  Optional[bytes] = None
    reuse_port: bool            = False
    
    def __post_init__(self):
        self.allow_reuse_port = self.reuse_port
        # build handler for base init
        self.server.__init__(self, self.address, new_handler( #pyright: ignore
            base=self.handler, 
            factory=self.factory, 
            args=self.args, 
            kwargs=self.kwargs,
            timeout=self.timeout,
            interface=self.interface,
            blocksize=getattr(self, 'blocksize', 8192),
        ))
    
    def __exit__(self, *_):
        """ensure server is shutdown properly"""
        self.shutdown()

    def shutdown(self):
        raise NotImplementedError

@dataclass
class UdpThreadServer(BaseThreadServer, socketserver.UDPServer):
    server  = socketserver.UDPServer
    handler = UdpHandler

    allow_broadcast: bool = False

    def __post_init__(self):
        super().__post_init__()
        if self.allow_broadcast:
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    
    def shutdown(self):
        """override shutdown behavior"""
        self.server.shutdown(self)
        self.socket.close()
        self.server_close()

@dataclass
class TcpThreadServer(BaseThreadServer, socketserver.TCPServer):
    server  = socketserver.TCPServer
    handler = TcpHandler

    ssl:       Optional[SSLContext] = None
    blocksize: int                  = 8192
    
    def __post_init__(self):
        super().__post_init__()
        if self.ssl:
            self.socket = wrap_socket(self.socket, server_side=True)

    def shutdown(self):
        """override shutdown behavior"""
        self.socket: socket.socket
        self.server.shutdown(self) #pyright: ignore
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()
        self.server_close()




from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    readme = f.read()

setup(
    name='pyserve3',
    version='0.0.2',
    license='MIT',
    packages=find_packages(),
    url='https://github.com/imgurbot12/pyserve',
    author='Andrew Scott',
    author_email='imgurbot12@gmail.com',
    description='Unify SocketServer Implementations based on a Session Model',
    python_requires='>=3.7',
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



"""
AsyncIO Implementations of session-based servers
"""
import asyncio
from dataclasses import dataclass, field
from typing import Type, Optional, Dict, Any

from .abc import RawAddr, Address, Writer, Session, modify_socket

#** Variables **#
__all__ = ['UdpProtocol', 'TcpProtocol']

#** Classes **#

@dataclass
class UdpWriter(Writer):
    addr:      Address
    transport: asyncio.DatagramTransport
    
    def write(self, data: bytes, addr: Optional[Address] = None):
        self.transport.sendto(data, addr or self.addr)

    def close(self):
        self.transport.close()

    def is_closing(self) -> bool:
        return self.transport.is_closing()

@dataclass
class BaseProtocol:
    factory:   Type[Session]
    args:      tuple           = field(default_factory=tuple)
    kwargs:    Dict[str, Any]  = field(default_factory=dict)
    timeout:   Optional[int]   = None
    interface: Optional[bytes] = None

    def test_factory(self):
        """validate session can be generated w/ args and kwargs"""
        self.factory(*self.args, **self.kwargs)

    def set_timeout(self, timeout: int, tport: asyncio.Transport):
        """implement timeout handling for async servers"""
        # generate async expiration function
        async def expire():
            await asyncio.sleep(timeout)
            if not tport.is_closing():
                tport.abort()
        # spawn future to kill transport
        loop = asyncio.get_event_loop()
        loop.create_task(expire())

class UdpProtocol(BaseProtocol, asyncio.DatagramTransport):
 
    def connection_made(self, transport: asyncio.DatagramTransport):
        """"""
        self.transport = transport
        # handle modifying socket and making changes for handling
        socket = transport.get_extra_info('socket')
        modify_socket(socket, None, self.interface)
        if self.timeout is not None:
            self.set_timeout(self.timeout, transport) #pyright: ignore
       
    def datagram_received(self, data: bytes, addr: RawAddr):
        """"""
        # generate session w/ attributes and notify on connection-made
        address      = Address(*addr)
        writer       = UdpWriter(address, self.transport)
        self.session = self.factory(*self.args, **self.kwargs)
        self.session.connection_made(address, writer)
        self.session.data_recieved(data)

    def connection_lost(self, err: Optional[Exception]):
        """"""
        self.session.connection_lost(err)

class TcpProtocol(BaseProtocol, asyncio.Protocol):
   
    def connection_made(self, transport: asyncio.Transport):
        """"""
        # collect attributes and prepare socket for session
        address = Address(*transport.get_extra_info('peername'))
        socket  = transport.get_extra_info('socket')
        # handle modifying socket and making changes for handling
        modify_socket(socket, None, self.interface)
        if self.timeout is not None:
            self.set_timeout(self.timeout, transport)
        # generate session w/ attributes and notify on connection-made
        self.session = self.factory(*self.args, **self.kwargs)
        self.session.connection_made(address, transport)

    def data_received(self, data: bytes):
        """"""
        self.session.data_recieved(data)

    def connection_lost(self, err: Optional[Exception]):
        """"""
        self.session.connection_lost(err)
